import pathlib

import pandas as pd


def extract_strand_merged_cg(output_dir_path, chrom_size_path, mc_context='CGN', cpu=5):
    output_dir = pathlib.Path(output_dir_path).absolute()
    group_table_path = output_dir / 'GROUP_TABLE.csv'

    group_table = pd.read_csv(group_table_path, index_col=0)
    group_levels = group_table.columns[1:]
    qsub_dir = output_dir / 'qsub'

    commands = []
    for group_level in group_levels:
        file_records = []
        group_dir = output_dir / group_level
        group_allc_records = group_dir / 'ALLC_table.csv'
        group_allc_df = pd.read_csv(group_allc_records, index_col=0)

        for allc_path in group_allc_df['AllcPath']:
            output_prefix = str(allc_path[:-12])
            output_path = output_prefix + '.CGN-Merge.allc.tsv.gz'
            cmd = f'allcools extract-allc ' \
                  f'--allc_path {allc_path} ' \
                  f'--output_prefix {output_prefix} ' \
                  f'--mc_contexts {mc_context} ' \
                  f'--chrom_size_path {chrom_size_path} ' \
                  f'--strandness merge ' \
                  f'--output_format allc ' \
                  f'--cov_cutoff 99999 ' \
                  f'--cpu {cpu}'
            commands.append(cmd)
            file_records.append(output_path)
        # group_allc_records will be updated
        group_allc_df['mCGExtract'] = file_records
        group_allc_df.to_csv(group_allc_records)

    command_file_path = qsub_dir / '_extract_cg_commands.txt'
    with open(command_file_path, 'w') as f:
        f.write('\n'.join(commands))
    return command_file_path



"""
Generate raw FASTQ dataframe based on fixed name pattern
name pattern is based on samplesheet generated in plateinfo_and_samplesheet.py
"""

import glob
import logging
import pathlib

import pandas as pd

# logger
log = logging.getLogger()


def _parse_v1_fastq_path(path):
    """
    UID pattern of V1 {sample_id_prefix}-{plate1}-{plate2}-{plate_pos}
    FASTQ name pattern of V1:
    {sample_id_prefix}-{plate1}-{plate2}-{plate_pos}_{internal_info}_{lane}_{read_type}_{internal_info}.fastq.gz
    """
    path = pathlib.Path(path)
    try:
        *_, plate1, plate2, multi_field = path.name.split('-')
        plate_pos, _, lane, read_type, _ = multi_field.split('_')
        try:
            assert plate_pos[0] in 'ABCDEFGH'
            assert int(plate_pos[1:]) in list(range(1, 13))
            assert lane in {'L001', 'L002', 'L003', 'L004'}
            assert read_type in {'R1', 'R2'}
            assert plate1 != plate2
        except AssertionError:
            raise ValueError
    except ValueError:
        raise ValueError(f'Found unknown name pattern in path {path}')
    name_dict = dict(plate1=plate1,
                     plate2=plate2,
                     plate_pos=plate_pos,
                     lane=lane,
                     read_type=read_type,
                     fastq_path=path,
                     uid=f'{plate1}-{plate2}-{plate_pos}')
    name_series = pd.Series(name_dict)
    return name_series


def _parse_v2_fastq_path(path):
    """
    UID pattern of V2 {sample_id_prefix}-{plate}-{multiplex_group}-{barcode_name}
    FASTQ name pattern of V1:
    {sample_id_prefix}-{plate}-{multiplex_group}-{barcode_name}_{internal_info}_{lane}_{read_type}_{internal_info}.fastq.gz
    """
    path = pathlib.Path(path)
    try:
        *_, plate, multiplex_group, multi_field = path.name.split('-')
        primer_name, _, lane, read_type, _ = multi_field.split('_')
        try:
            assert primer_name[0] in 'ABCDEFGHIJKLMNOP'
            assert int(primer_name[1:]) in list(range(1, 25))
            assert int(multiplex_group) in list(range(1, 7))
            assert lane in {'L001', 'L002', 'L003', 'L004'}
            assert read_type in {'R1', 'R2'}
        except AssertionError:
            raise ValueError
    except ValueError:
        raise ValueError(f'Found unknown name pattern in path {path}')
    name_dict = dict(plate=plate,
                     multiplex_group=multiplex_group,
                     primer_name=primer_name,
                     lane=lane,
                     read_type=read_type,
                     fastq_path=path,
                     uid=f'{plate}-{multiplex_group}-{primer_name}')
    name_series = pd.Series(name_dict)
    return name_series


def make_fastq_dataframe(file_path, barcode_version, output_path=None):
    """
    Generate fastq_dataframe for pipeline input.

    Parameters
    ----------
    file_path
        Accept 1. path pattern contain wildcard, 2. path list, 3. path of one file contain all the paths.
    barcode_version
        Only accept two options: 1) V1 for 8 random index; 2) V2 for 384 random index.
    output_path
        output path of the fastq dataframe
    Returns
    -------
        fastq_dataframe for pipeline input.
    """
    barcode_version = barcode_version.upper()
    if barcode_version == 'V1':
        parser = _parse_v1_fastq_path
    elif barcode_version == 'V2':
        parser = _parse_v2_fastq_path
    else:
        raise ValueError(f'Primer Version can only be V1 or V2, got {barcode_version}.')

    if isinstance(file_path, str) and ('*' in file_path):
        file_path = [str(pathlib.Path(p).absolute()) for p in glob.glob(file_path)]
    elif isinstance(file_path, list):
        pass
    else:
        with open(file_path) as f:
            file_path = [line.strip() for line in f]
    log.info(f'{len(file_path)} FASTQ file paths in input')

    fastq_data = []
    for path in file_path:
        name_series = parser(path)
        fastq_data.append(name_series)
    fastq_df = pd.DataFrame(fastq_data)
    log.info(f'{fastq_df.shape[0]} valid fastq names.')
    if fastq_df.shape[0] == 0:
        log.info('No fastq name remained, check if the name pattern is correct.')
        return None

    # make sure UID is unique
    for _, df in fastq_df.groupby(['lane', 'read_type']):
        if df['uid'].unique().size != df['uid'].size:
            raise ValueError(f'UID column is not unique.')
    if output_path is not None:
        fastq_df.to_csv(output_path, index=False)
    return fastq_df



import pathlib

import pandas as pd

from .utilities import parse_trim_fastq_stats, parse_trim_fastq_stats_mct, \
    parse_bismark_report, parse_deduplicate_stat, \
    generate_allc_stats


def mc_mapping_stats(output_dir, config):
    """this may apply to single UID dir, so config is provided as parameter"""
    output_dir = pathlib.Path(output_dir).absolute()
    fastq_dir = output_dir / 'fastq'
    bam_dir = output_dir / 'bam'
    allc_dir = output_dir / 'allc'
    cell_stats = []
    cell_ids = [path.name.split('.')[0]
                for path in allc_dir.glob(f'*.allc.tsv.gz')]

    for cell_id in cell_ids:
        print(f'Parsing stats of {cell_id}.')
        total_stats = []
        for read_type in ['R1', 'R2']:
            if config['mode'] == 'mct':
                total_stats.append(
                    parse_trim_fastq_stats_mct(
                        fastq_dir / f'{cell_id}-{read_type}.trimmed.stats.txt'))
            else:
                total_stats.append(
                    parse_trim_fastq_stats(
                        fastq_dir / f'{cell_id}-{read_type}.trimmed.stats.tsv'))
            total_stats.append(
                parse_bismark_report(
                    bam_dir / f'{cell_id}-{read_type}.trimmed_bismark_bt2_SE_report.txt'))
            total_stats.append(
                parse_deduplicate_stat(
                    bam_dir / f'{cell_id}-{read_type}.trimmed_bismark_bt2.deduped.matrix.txt'
                ))
        cell_stats.append(pd.concat(total_stats))
    mapping_df = pd.DataFrame(cell_stats)
    mapping_df.index.name = 'cell_id'

    # add allc stats
    allc_df = generate_allc_stats(output_dir, config)
    final_df = pd.concat([mapping_df, allc_df], sort=True, axis=1)
    return final_df


def mc_additional_cols(final_df):
    """Additional columns for mC mapping summary"""
    final_df = final_df.copy()
    final_df['CellInputReadPairs'] = final_df['R1InputReads'].astype(int)  # == final_df['R2InputReads']
    if 'PCRIndex' in final_df.columns:  # plate info might not exist if the cell name is abnormal
        cell_barcode_ratio = pd.concat([(i['CellInputReadPairs'] / i['CellInputReadPairs'].sum())
                                        for _, i in final_df.groupby('PCRIndex')])
        final_df['CellBarcodeRatio'] = cell_barcode_ratio

    final_df['FinalmCReads'] = final_df['R1FinalBismarkReads'] + final_df['R2FinalBismarkReads']
    return final_df



import pathlib

import pandas as pd
from pysam import AlignmentFile

from .utilities import parse_trim_fastq_stats, generate_allc_stats


def _m3c_bam_unique_read_counts(bam_path, read_type_int):
    unique_reads = set()
    with AlignmentFile(bam_path) as bam:
        for read in bam:
            unique_reads.add(read.query_name.split(f'_{read_type_int}:N:0:')[0])
    return len(unique_reads)


def _m3c_count_bams(bam_dir, cell_id, read_type):
    bam_path_dict = {
        f'{read_type}UniqueMappedReads': bam_dir / f'{cell_id}-{read_type}.two_mapping.filter.bam',
        f'{read_type}DeduppedReads': bam_dir / f'{cell_id}-{read_type}.two_mapping.deduped.bam',
    }
    read_counts = {name: _m3c_bam_unique_read_counts(path, 1 if read_type == 'R1' else 2)
                   for name, path in bam_path_dict.items()}
    return pd.Series(read_counts, name=cell_id)


def m3c_mapping_stats(output_dir, config):
    """this may apply to single UID dir, so config is provided as parameter"""
    output_dir = pathlib.Path(output_dir).absolute()
    fastq_dir = output_dir / 'fastq'
    bam_dir = output_dir / 'bam'
    hic_dir = output_dir / 'hic'
    cell_stats = []
    cell_ids = [path.name.split('.')[0]
                for path in bam_dir.glob('*.3C.sorted.bam')]

    for cell_id in cell_ids:
        total_stats = []  # list of series
        for read_type in ['R1', 'R2']:
            # fastq reads
            total_stats.append(
                parse_trim_fastq_stats(
                    fastq_dir / f'{cell_id}-{read_type}.trimmed.stats.tsv'))
            # bam reads
            total_stats.append(
                _m3c_count_bams(bam_dir, cell_id, read_type)
            )
        # contacts
        contact_counts = pd.read_csv(hic_dir / f'{cell_id}.3C.contact.tsv.gz.counts.txt',
                                     header=None, index_col=0, squeeze=True)
        contact_counts.name = cell_id
        total_stats.append(contact_counts)

        cell_stats.append(pd.concat(total_stats))
    total_df = pd.DataFrame(cell_stats)

    # add allc stats
    allc_df = generate_allc_stats(output_dir, config)
    final_df = pd.concat([total_df, allc_df], sort=True, axis=1)
    return final_df


def m3c_additional_cols(final_df):
    final_df['FinalmCReads'] = final_df['R1DeduppedReads'] + final_df['R2DeduppedReads']
    final_df['CellInputReadPairs'] = final_df['R1InputReads']
    # use % to be consistent with others
    final_df['R1MappingRate'] = final_df['R1UniqueMappedReads'] / final_df['R1TrimmedReads'] * 100
    final_df['R2MappingRate'] = final_df['R2UniqueMappedReads'] / final_df['R2TrimmedReads'] * 100
    final_df['R1DuplicationRate'] = (1 - final_df['R1DeduppedReads'] / final_df['R1UniqueMappedReads']) * 100
    final_df['R2DuplicationRate'] = (1 - final_df['R2DeduppedReads'] / final_df['R2UniqueMappedReads']) * 100

    if 'PCRIndex' in final_df.columns:  # plate info might not exist if the cell name is abnormal
        cell_barcode_ratio = pd.concat([(i['CellInputReadPairs'] / i['CellInputReadPairs'].sum())
                                        for _, i in final_df.groupby('PCRIndex')])
        final_df['CellBarcodeRatio'] = cell_barcode_ratio

    final_df['TotalContacts'] = final_df[
        ['CisShortContact', 'CisLongContact', 'TransContact']].sum(axis=1)
    final_df['CisShortRatio'] = final_df['CisShortContact'] / final_df['TotalContacts']
    final_df['CisLongRatio'] = final_df['CisLongContact'] / final_df['TotalContacts']
    final_df['TransRatio'] = final_df['TransContact'] / final_df['TotalContacts']
    return final_df



import pathlib

import pandas as pd

from .allc_to_bigwig import generate_bigwig
from .extract_allc import extract_strand_merged_cg
from .merge_allc import _merge_cell, _merge_cluster


def bulk_pipeline(
        output_dir_path,
        group_table_path,
        chrom_size_path,
        binarize_single_cell=False,
        merge_cpu=10,
        ignore_names=None,
        max_cell_group=999,
        cg_context='CGN',
        bigwig_context=None,
        bigwig_binsize=50):
    output_dir = pathlib.Path(output_dir_path)
    output_dir.mkdir(exist_ok=True, parents=True)
    qsub_dir = output_dir / 'qsub'
    qsub_dir.mkdir(exist_ok=True)

    group_table = pd.read_csv(group_table_path, index_col=0).fillna('NOT_A_CLUSTER')
    group_table_path = output_dir / 'GROUP_TABLE.csv'
    group_table.to_csv(group_table_path)

    # merge ALLC
    # step 1: merge cell ALLC
    command_file_list = []
    cell_group_allc_map, command_file_path = _merge_cell(group_table_path,
                                                         output_dir_path,
                                                         chrom_size_path,
                                                         binarize=binarize_single_cell,
                                                         cpu=merge_cpu,
                                                         max_cell_group=max_cell_group)
    command_file_list.append(command_file_path)

    # step 2: merge cluster ALLC
    if ignore_names is None:
        ignore_names = ['NOT_A_CLUSTER']
    else:
        ignore_names.append('NOT_A_CLUSTER')

    command_file_path = _merge_cluster(output_dir_path,
                                       cell_group_allc_map,
                                       ignore_names,
                                       chrom_size_path,
                                       cpu=merge_cpu)
    command_file_list.append(command_file_path)

    # step3: extract mCG ALLC
    command_file_path = extract_strand_merged_cg(output_dir_path,
                                                 chrom_size_path,
                                                 mc_context=cg_context,
                                                 cpu=1)
    command_file_list.append(command_file_path)

    # step4: bigwig
    if bigwig_context is not None:
        command_file_path = generate_bigwig(output_dir_path,
                                            chrom_size_path,
                                            cpu=1,
                                            mc_contexts_list=bigwig_context,
                                            bin_size=bigwig_binsize)
        command_file_list.append(command_file_path)
    return



"""
When adding new function:
1. add a func_register_subparser function to register the subparser
2. add a condition in main func about this new func name, import the real func as func in main
"""

import argparse
import logging
import sys

import cemba_data
from cemba_data import __version__

log = logging.getLogger()

DESCRIPTION = """
YAP (Yet Another Pipeline) is a mapping pipeline for multiple
snmC-seq based single-cell sequencing technologies.

See documentation at https://hq-1.gitbook.io/mc/
"""

EPILOG = """
Author: Hanqing Liu, hanliu@salk.edu

If this looks good, send coffee to...
"""


class NiceFormatter(logging.Formatter):
    """
    From Cutadapt https://github.com/marcelm/cutadapt
    Do not prefix "INFO:" to info-level log messages (but do it for all other
    levels).
    Based on http://stackoverflow.com/a/9218261/715090 .
    """

    def format(self, record):
        if record.levelno != logging.INFO:
            record.msg = '{}: {}'.format(record.levelname, record.msg)
        return super().format(record)


def setup_logging(stdout=False, quiet=False, debug=False):
    """
    From Cutadapt https://github.com/marcelm/cutadapt
    Attach handler to the global logger object
    """
    # Due to backwards compatibility, logging output is sent to standard output
    # instead of standard error if the -o option is used.
    stream_handler = logging.StreamHandler(sys.stdout if stdout else sys.stderr)
    stream_handler.setFormatter(NiceFormatter())
    # debug overrides quiet
    if debug:
        level = logging.DEBUG
    elif quiet:
        level = logging.ERROR
    else:
        level = logging.INFO
    stream_handler.setLevel(level)
    log.setLevel(level)
    log.addHandler(stream_handler)


def qsub_register_subparser(subparser):
    parser = subparser.add_parser('qsub',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="General qsub helper, need to prepare a command dict file")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--command_file_path",
        type=str,
        required=True,
        nargs='+',
        help="One or space-separated paths of the command file. Accept 2 different command file format: "
             "1. each line in the file is a full command, will be submitted as a single job. "
             "Qsub parameters can only be specified by qsub_global_parms in this way."
             "2. JSON format, a list of dict, where each dict is a command (required key \"command\") "
             "and other optional key specify qsub parameters per job."
    )

    parser_req.add_argument(
        "--working_dir",
        type=str,
        required=True,
        help="Working directory of the qsub project"
    )

    parser_req.add_argument(
        "--project_name",
        type=str,
        required=True,
        help="Name of the qsub project"
    )

    parser.add_argument(
        "--wait_until",
        type=str,
        nargs='+',
        required=False,
        help="If provided with a space separate job id(s), "
             "this job will wait until those job finish first."
    )

    parser.add_argument(
        "--total_cpu",
        type=int,
        required=False,
        default=30,
        help="Total concurrent CPU in qsub list, together with total_mem, "
             "determines how many jobs running in parallel."
    )

    parser.add_argument(
        "--total_mem",
        type=int,
        required=False,
        default=9999,
        help="Total concurrent MEM (GBs) in qsub list, together with total_cpu, "
             "determines how many jobs running in parallel. "
             "Default is 9999, which means only use total_cpu to determine."
    )

    parser.add_argument(
        "--qsub_global_parms",
        type=str,
        required=False,
        default='',
        help="Other global qsub parameters you want to pass to each job's qsub script. "
             "This will cover command.json if set repeatedly."
             "These additional parameters should form in one ';' separated string like this: "
             "'-q=queue_name;-cwd;-wd=/path/to/working/dir;-pe smp=20;-l h_vmem=5G'"
    )
    return


def sbatch_register_subparser(subparser):
    parser = subparser.add_parser('sbatch',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="General sbatch helper, "
                                       "need to prepare a command dict file where key is a command, "
                                       "value is a list of target files.")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--project_name",
        type=str,
        required=True,
        help="Name of the qsub project"
    )

    parser_req.add_argument(
        "--command_file_path",
        type=str,
        required=True,
        help="Each row is a command."
    )

    parser_req.add_argument(
        "--working_dir",
        type=str,
        required=True,
        help="Working directory of the qsub project"
    )

    parser_req.add_argument(
        "--time_str",
        type=str,
        required=True,
        help="Time estimate (upper limit) of Sbatch Jobs."
    )

    parser_req.add_argument(
        "--queue",
        type=str,
        default='skx-normal',
        choices=['skx-normal'],
        help="Queue partition of stampede2. Right now only support skx-normal, "
             "which is large memory nodes suitable for mapping jobs. "
    )

    parser_req.add_argument(
        "--email",
        type=str,
        default=None,
        help="If you want to get notice about job status, put your email here."
    )

    parser_req.add_argument(
        "--email_type",
        type=str,
        default='fail',
        help="Type of status you want to get notice, default is fail, means only get notice when job failed."
    )

    parser.add_argument(
        "--max_jobs",
        type=int,
        required=False,
        default=None,
        help="Max number of jobs for the same user (not for the same sbatch command). "
             "If not provided, will determine automatically based on stampede2 limits."
    )

    parser.add_argument(
        "--dry_run",
        required=False,
        action='store_true',
        help="Prepare scripts for sbatch manual submission or debug."
    )
    return


def print_plate_info_register_subparser(subparser):
    parser = subparser.add_parser('default-plate-info',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Print out default plate info template.")

    parser.add_argument(
        "--barcode_version", '-v', '-V',
        type=str,
        default='V2',
        choices=['V1', 'V2'],
        help="Use V1 template for 8-random-index library version, "
             "Use V2 template for 384-random-index library version."
    )
    return


def make_sample_sheet_register_subparser(subparser):
    parser = subparser.add_parser('make-sample-sheet',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Print out default plate info template.")

    parser_req = parser.add_argument_group("Required inputs")
    parser_opt = parser.add_argument_group("Optional inputs")

    parser_req.add_argument(
        "--plate_info_path",
        "-p",
        type=str,
        required=True,
        help="Path of the plate information file."
    )

    parser_req.add_argument(
        "--output_prefix",
        "-o",
        type=str,
        required=True,
        help="Output prefix, will generate 2 sample sheets, 1 for miseq, 1 for novaseq"
    )

    parser_opt.add_argument(
        "--header_path",
        type=str,
        help="Path to the sample sheet header that contains sequencer info. Will use default if not provided."
    )

    return


def demultiplex_register_subparser(subparser):
    parser = subparser.add_parser('demultiplex',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Demultiplex bcl2fastq results.")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--fastq_pattern",
        "-fq",
        type=str,
        required=True,
        help="FASTQ files with wildcard to match all bcl2fastq results, pattern with wildcard must be quoted."
    )

    parser_req.add_argument(
        "--output_dir",
        "-o",
        type=str,
        required=True,
        help="Pipeline output directory, will be created recursively."
    )

    parser_req.add_argument(
        "--config_path",
        "-config",
        type=str,
        required=True,
        help="Path to the mapping config, see 'yap default-mapping-config' about how to generate this file."
    )

    parser_req.add_argument(
        "--cpu",
        '-j',
        type=int,
        required=True,
        help="Number of cores to use. Note that the demultiplex step will only use at most 32 cores, "
             "the merge lane step will use the number of cores you provided."
    )
    return


def print_default_config_register_subparser(subparser):
    parser = subparser.add_parser('default-mapping-config',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Print out default config of mapping pipeline")
    from cemba_data.utilities import MAPPING_MODE_CHOICES
    parser.add_argument(
        "--mode",
        type=str,
        required=True,
        choices=MAPPING_MODE_CHOICES,
        help="Library mode"
    )

    parser.add_argument(
        "--barcode_version",
        '-v',
        '-V',
        type=str,
        required=True,
        choices=['V1', 'V2'],
        help="Barcode version, V1 for 8 random index, V2 for 384 random index"
    )

    parser.add_argument(
        "--bismark_ref",
        type=str,
        required=True,
        help="Path to the bismark reference"
    )

    parser.add_argument(
        "--genome_fasta",
        type=str,
        required=True,
        help="Path to the genome fasta file"
    )

    parser.add_argument(
        "--star_ref",
        type=str,
        required=False,
        help="[mct only] Path to the STAR reference, required if mode is mct"
    )

    parser.add_argument(
        "--gtf",
        type=str,
        required=False,
        help="[mct only] Path to the GTF annotation file, required if mode is mct"
    )

    parser.add_argument(
        "--chrom_size_path",
        type=str,
        required=False,
        help="[m3c only] Path to the chrom size file, "
             "only chromosomes occur in this file will be considered in generating chromatin contacts, "
             "required if mode is mct"
    )

    parser.add_argument(
        "--nome",
        dest='nome',
        action='store_true',
        help='Does this library have NOMe treatment?'
    )
    parser.set_defaults(nome=False)
    return


def start_from_cell_fastq_register_subparser(subparser):
    parser = subparser.add_parser('start-from-cell-fastq',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Prepare batch jobs directly from cell-level FASTQ files, "
                                       "no demultiplex step.")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--output_dir",
        '-o',
        type=str,
        required=True,
        help="Pipeline output directory, if not exist, will create recursively."
    )

    parser_req.add_argument(
        "--config_path",
        "-config",
        type=str,
        required=True,
        help="Path to the mapping config, see 'yap default-mapping-config' about how to generate this file."
    )

    parser_req.add_argument(
        "--fastq_pattern",
        "-fq",
        type=str,
        required=True,
        help="Path pattern with wildcard to match all cell-level FASTQ files, pattern with wildcard must be quoted."
    )
    return


def summary_register_subparser(subparser):
    parser = subparser.add_parser('summary',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Mapping summary after the pipeline finished.")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--output_dir",
        '-o',
        type=str,
        required=True,
        help="Pipeline output directory."
    )

    parser_req.add_argument(
        "--notebook",
        '-nb',
        type=str,
        required=False,
        default=None,
        help="Notebook template for mapping summary, if not provided, will use yap default template."
    )
    return


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION,
                                     epilog=EPILOG,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("--version", action="version", help="Show version number and exit",
                        version=__version__)
    subparsers = parser.add_subparsers(
        title="functions",
        dest="command",
        metavar=""
    )

    # add subparsers
    qsub_register_subparser(subparsers)
    sbatch_register_subparser(subparsers)
    print_default_config_register_subparser(subparsers)
    print_plate_info_register_subparser(subparsers)
    make_sample_sheet_register_subparser(subparsers)
    demultiplex_register_subparser(subparsers)
    start_from_cell_fastq_register_subparser(subparsers)
    summary_register_subparser(subparsers)

    # initiate
    args = None
    if len(sys.argv) > 1:
        # print out version
        if sys.argv[1] in ['-v', '--version']:
            print(cemba_data.__version__)
            exit()
        else:
            args = parser.parse_args()
    else:
        # print out help
        parser.parse_args(["-h"])
        exit()

    # set up logging
    if not logging.root.handlers:
        setup_logging(stdout=True,
                      quiet=False)

    # execute command
    args_vars = vars(args)
    cur_command = args_vars.pop('command')
    # Do real import here:
    if cur_command == 'qsub':
        from cemba_data.qsub import qsub as func
    elif cur_command == 'sbatch':
        from cemba_data.sbatch import sbatch_submitter as func
    elif cur_command == 'default-plate-info':
        from .demultiplex import print_plate_info as func
    elif cur_command == 'make-sample-sheet':
        from .demultiplex import make_sample_sheet as func
    elif cur_command == 'demultiplex':
        from .demultiplex import demultiplex_pipeline as func
    elif cur_command == 'default-mapping-config':
        from .mapping import print_default_mapping_config as func
    elif cur_command == 'start-from-cell-fastq':
        from .mapping import start_from_cell_fastq as func
    elif cur_command == 'summary':
        from cemba_data.mapping import final_summary as func
    else:
        log.debug(f'{cur_command} not Known, check the main function if else part')
        parser.parse_args(["-h"])
        return

    # run the command
    func(**args_vars)
    return


if __name__ == '__main__':
    main()



import pathlib

import cemba_data
from ..utilities import MAPPING_MODE_CHOICES

# Load defaults
PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])


def print_default_mapping_config(mode, barcode_version, bismark_ref, genome_fasta,
                                 star_ref=None, gtf=None, nome=False, chrom_size_path=None):
    mode = mode.lower()
    if mode not in MAPPING_MODE_CHOICES:
        raise ValueError(f'Unknown mode {mode}')

    barcode_version = barcode_version.upper()
    if barcode_version not in ['V1', 'V2']:
        raise ValueError(f'Unknown mode {barcode_version}')

    bismark_ref = pathlib.Path(bismark_ref).absolute()

    if mode == 'mct':
        if (star_ref is None) or (gtf is None):
            raise ValueError('star_ref and gtf must be provided when mode is mct.')
        star_ref = pathlib.Path(star_ref).absolute()
        gtf = pathlib.Path(gtf).absolute()
    if mode == 'm3c':
        if chrom_size_path is None:
            raise ValueError('chrom_size_path must be provided when mode is m3c.')

    genome_fasta = pathlib.Path(genome_fasta).absolute()

    if mode == 'mc':
        if nome:
            config_path = PACKAGE_DIR / 'files/default_config/mapping_config_nome.ini'
        else:
            config_path = PACKAGE_DIR / 'files/default_config/mapping_config_mc.ini'
        with open(config_path) as f:
            config_content = f.read()
    elif mode == 'mct':
        if nome:
            config_path = PACKAGE_DIR / 'files/default_config/mapping_config_mct-nome.ini'
        else:
            config_path = PACKAGE_DIR / 'files/default_config/mapping_config_mct.ini'
        with open(config_path) as f:
            config_content = f.read()
        config_content = config_content.replace('CHANGE_THIS_TO_YOUR_STAR_REFERENCE_DIR', str(star_ref))
        config_content = config_content.replace('CHANGE_THIS_TO_YOUR_GENE_ANNOTATION_GTF', str(gtf))
    elif mode == 'm3c':
        config_path = PACKAGE_DIR / 'files/default_config/mapping_config_m3c.ini'
        with open(config_path) as f:
            config_content = f.read()
        config_content = config_content.replace('CHANGE_THIS_TO_YOUR_CHROM_SIZE_PATH', str(chrom_size_path))
    else:
        raise

    config_content = config_content.replace('USE_CORRECT_BARCODE_VERSION_HERE', barcode_version)
    config_content = config_content.replace('CHANGE_THIS_TO_YOUR_BISMARK_REFERENCE_DIR', str(bismark_ref))
    config_content = config_content.replace('CHANGE_THIS_TO_YOUR_REFERENCE_FASTA', str(genome_fasta))
    print(config_content)
    return



import json
import pathlib
from collections import defaultdict

import pandas as pd


def _merge_cell(group_table_path, output_dir_path, chrom_size_path, binarize, cpu, max_cell_group):
    output_dir = pathlib.Path(output_dir_path).absolute()
    qsub_dir = output_dir / 'qsub'

    group_table = pd.read_csv(group_table_path, index_col=0)

    if binarize:
        binarize_param = '--binarize'
    else:
        binarize_param = ''

    # cell level merge
    cell_merge_output_dir = output_dir / 'CELL_GROUP'
    cell_merge_output_dir.mkdir(exist_ok=True)

    command_records = {}
    col_names = group_table.columns[1:].tolist()
    file_map = {col_name: defaultdict(list) for col_name in col_names}
    for i, (cluster_combination, sub_df) in enumerate(group_table.groupby(col_names)):
        cell_number = sub_df.shape[0]
        if (max_cell_group is not None) and (cell_number > max_cell_group):
            sub_df = sub_df.sample(max_cell_group, random_state=1)
            cell_number = max_cell_group

        output_path = cell_merge_output_dir / f'{i}-{cell_number}.allc.tsv.gz'
        file_list_path = cell_merge_output_dir / f'{i}-{cell_number}.cell_list.txt'

        if isinstance(cluster_combination, str):
            cluster_combination = [cluster_combination]

        for col_name, cluster_name in zip(col_names, cluster_combination):
            file_map[col_name][cluster_name].append(str(output_path))

        # dump file list
        with open(file_list_path, 'w') as f:
            f.write('\n'.join(sub_df['AllcPath']))

        # command
        if cell_number == 1:
            input_path = sub_df['AllcPath'][0]
            cmd = f'cp {input_path} {output_path}; ' \
                  f'cp {input_path}.tbi {output_path}.tbi'
        else:
            input_path = file_list_path
            cmd = f'allcools merge-allc ' \
                  f'--allc_paths {input_path} ' \
                  f'--output_path {output_path} ' \
                  f'--chrom_size_path {chrom_size_path} ' \
                  f'--cpu {cpu} {binarize_param}'
        command_records[cmd] = cell_number

    command_path = qsub_dir / '_merge_cell_commands.txt'
    with open(command_path, 'w') as f:
        # write down commands order by cell number, larger number first
        for cmd in pd.Series(command_records).sort_values(ascending=False).index:
            f.write(cmd + '\n')

    with open(cell_merge_output_dir / '_file_map.json', 'w') as f:
        json.dump(file_map, f)
    return file_map, command_path


def _merge_cluster(output_dir_path, cell_group_allc_map, ignore_names, chrom_size_path, cpu):
    output_dir = pathlib.Path(output_dir_path).absolute()
    qsub_dir = output_dir / 'qsub'

    if ignore_names is None:
        ignore_names = []

    command_records = []
    for col, cluster_path_map in cell_group_allc_map.items():
        col_output_dir = output_dir / col
        col_output_dir.mkdir(exist_ok=True)
        file_path_records = []
        for cluster, cell_group_allc_list in cluster_path_map.items():
            if cluster in ignore_names:
                continue

            output_path = col_output_dir / f'{col}.{cluster}.allc.tsv.gz'

            total_cell_number = 0
            for cell_group_allc_path in cell_group_allc_list:
                # parse cell number from cell_group_allc file name, fix by _merge_cell_allc
                total_cell_number += int(pathlib.Path(cell_group_allc_path).name.split('.')[0].split('-')[-1])

            if len(cell_group_allc_list) == 1:
                cmd = f'cp {cell_group_allc_list[0]} {output_path}; ' \
                      f'cp {cell_group_allc_list[0]}.tbi {output_path}.tbi'
            else:
                input_path = col_output_dir / f'{col}.{cluster}.list.txt'
                with open(input_path, 'w') as f:
                    f.write('\n'.join(cell_group_allc_list))
                cmd = f'allcools merge-allc ' \
                      f'--allc_paths {input_path} ' \
                      f'--output_path {output_path} ' \
                      f'--chrom_size_path {chrom_size_path} ' \
                      f'--cpu {cpu}'
            command_records.append(cmd)
            file_path_records.append([col, cluster, output_path, total_cell_number])
        file_df = pd.DataFrame(file_path_records,
                               columns=['LevelName', 'ClusterName',
                                        'AllcPath', 'TotalCellCombined'])
        file_df.to_csv(col_output_dir / 'ALLC_table.csv', index=None)

    command_path = qsub_dir / '_merge_cluster_commands.txt'
    with open(command_path, 'w') as f:
        f.write('\n'.join(command_records))
    return command_path



import pathlib
import subprocess

import pandas as pd
from papermill import execute_notebook, PapermillExecutionError

from .m3c import m3c_mapping_stats, m3c_additional_cols
from .mc import mc_mapping_stats, mc_additional_cols
from .mct import mct_mapping_stats, mct_additional_cols
from .plate_info import get_plate_info
from ..pipelines import PACKAGE_DIR
from ...utilities import get_configuration


def mapping_stats(output_dir):
    """This is UID level mapping summary, the config file is in parent dir"""
    output_dir = pathlib.Path(output_dir).absolute()
    config = get_configuration(output_dir.parent / 'mapping_config.ini')
    mode = config['mode']

    if mode == 'mc':
        final_df = mc_mapping_stats(output_dir, config)
    elif mode == 'mct':
        final_df = mct_mapping_stats(output_dir, config)
    elif mode == 'm3c':
        final_df = m3c_mapping_stats(output_dir, config)
    else:
        raise ValueError

    # plate info, which is tech independent.
    _plate_info = get_plate_info(final_df.index, barcode_version=config['barcode_version'])
    final_df = pd.concat([_plate_info, final_df], axis=1)

    # save
    final_df.to_csv(output_dir / 'MappingSummary.csv.gz')
    return


def final_summary(output_dir, cleanup=True, notebook=None):
    output_dir = pathlib.Path(output_dir).absolute()
    mode = get_configuration(output_dir / 'mapping_config.ini')['mode']
    path_to_remove = []

    # Before running summary,
    # first make sure all the UID dir having Snakefile also has mapping summary (means successful)
    snakefile_list = list(output_dir.glob('*/Snakefile'))
    summary_paths = []
    missing_summary_dirs = []
    for path in snakefile_list:
        uid_dir = path.parent
        summary_path = uid_dir / 'MappingSummary.csv.gz'
        if summary_path.exists():
            summary_paths.append(summary_path)
        else:
            missing_summary_dirs.append(uid_dir)

    if len(missing_summary_dirs) != 0:
        print('These sub dir missing MappingSummary files:')
        for p in missing_summary_dirs:
            print(p)
        raise FileNotFoundError(f'Note that all sub dir should be successfully mapped '
                                f'before generating final summary. \n'
                                f'The MappingSummary.csv.gz is the final target file of snakefile in {path}. \n'
                                f'Run the corresponding snakemake command again to retry mapping.\n'
                                f'The snakemake commands can be found in output_dir/snakemake/*/snakemake_cmd.txt')

    # aggregate mapping summaries
    total_mapping_summary = pd.concat([pd.read_csv(path, index_col=0)
                                       for path in summary_paths])
    total_mapping_summary_path = output_dir / 'stats/MappingSummary.csv.gz'

    # if this is mct, aggregate all the gene counts
    if mode == 'mct':
        from ..stats.mct import aggregate_feature_counts
        aggregate_feature_counts(output_dir)

    # add additional columns based on some calculation
    if mode == 'mc':
        total_mapping_summary = mc_additional_cols(total_mapping_summary)
    elif mode == 'mct':
        total_mapping_summary = mct_additional_cols(total_mapping_summary, output_dir=output_dir)
    elif mode == 'm3c':
        total_mapping_summary = m3c_additional_cols(total_mapping_summary)
    else:
        raise

    # save total mapping summary
    total_mapping_summary.to_csv(total_mapping_summary_path)

    # add .snakemake files to deletion
    snakemake_hiding_dirs = list(output_dir.glob('*/.snakemake'))
    path_to_remove += snakemake_hiding_dirs

    # add temp dir in the bam dirs to deletion
    mapping_temp_dirs = list(output_dir.glob('*/bam/temp'))
    path_to_remove += mapping_temp_dirs

    # write a ALLC path file for generating MCDS
    allc_paths = pd.Series({path.name.split('.')[0]: str(path)
                            for path in output_dir.glob('*/allc/*tsv.gz')})
    allc_paths.to_csv(output_dir / 'stats/AllcPaths.tsv', sep='\t', header=False)

    if 'Plate' in total_mapping_summary.columns:  # only run notebook when plate info exist
        # run summary notebook
        nb_path = output_dir / 'stats/MappingSummary.ipynb'
        try:
            mode = get_configuration(output_dir / 'mapping_config.ini')['mode']
            if notebook is None:
                template_notebook = PACKAGE_DIR / f'files/mapping_summary_template/{mode}_template.ipynb'
            else:
                template_notebook = str(notebook)
            print(f'Using notebook template from {template_notebook}')
            print('Executing summary plotting notebook...')
            execute_notebook(
                input_path=str(template_notebook),
                output_path=str(nb_path),
                parameters=dict(output_dir=str(output_dir))
            )
            print('Summary notebook successfully executed. Exporting HTML...')
            subprocess.run(['jupyter', 'nbconvert', '--to', 'html', str(nb_path)])
            print(f'See the summary plots here: {str(nb_path)[:-5]}html')
            print(f'Or customize the summary plots here: {nb_path}')
        except PapermillExecutionError:
            print(f'Ops, summary plotting notebook got some error, check the information in {nb_path}')
            cleanup = False

    # delete
    if cleanup:
        print('Clean up snakemake log (might take several minutes) ...')
        for path in path_to_remove:
            subprocess.run(['rm', '-rf', str(path)], check=True)
    return
